/*
 * RingQueue.c
 *
 *  Created on: Oct 18, 2009
 *      Author: chwolfe2
 */

// size is always assumed to be 1.5GiB
#include "RingQueue.h"
#include <sys/mman.h>
#include <assert.h>
#include <string.h>

#define RING_QUEUE_STATIC_SIZE 1610612736
#define RING_QUEUE_REMOVAL_SIZE 262144
#define RING_QUEUE_OVERLAP_GRACE RING_QUEUE_REMOVAL_SIZE

StatusCode RingQueue_Create(RingQueue* rq, size_t insertionSize){
	assert(rq!=NULL);
	assert(rq->data==NULL);
	rq->data = mmap(NULL, RING_QUEUE_STATIC_SIZE + RING_QUEUE_OVERLAP_GRACE, PROT_READ | PROT_WRITE, MAP_PRIVATE	| MAP_ANONYMOUS | MAP_POPULATE | MAP_LOCKED, -1, 0);
	if (rq->data == MAP_FAILED) {
		return FAILURE;
	}
	rq->in = 0;
	rq->out = 0;
	rq->totalSize = RING_QUEUE_STATIC_SIZE;
	rq->bytesUsed = 0;
	rq->allocatedSize = RING_QUEUE_STATIC_SIZE + RING_QUEUE_OVERLAP_GRACE;
	rq->insertionSize = insertionSize;
	rq->allowUndersizeRemoval=false;
	return SUCCESS;
}
StatusCode RingQueue_Destroy(RingQueue* rq){
	assert(rq!=NULL);
	assert(rq->data != NULL);
	if ((munmap(rq->data, rq->allocatedSize)) == -1)
		return FAILURE;
	rq->in = 0;
	rq->out = 0;
	rq->totalSize = 0;
	rq->bytesUsed = 0;
	rq->allocatedSize = 0;
	rq->data=NULL;
	rq->insertionSize=0;
	return SUCCESS;
}
StatusCode RingQueue_Reset(RingQueue* rq){
	assert(rq!=NULL);
	assert(rq->data != NULL);
	rq->in = 0;
	rq->out = 0;
	rq->bytesUsed = 0;
	rq->allowUndersizeRemoval=false;
	return SUCCESS;
}
StatusCode RingQueue_NextInsertionPoint(RingQueue* rq, char** insertionPoint){
	if ((rq->totalSize - rq->bytesUsed) < rq->insertionSize){
		*insertionPoint = NULL;
		return FAILURE;
	} else {
		*insertionPoint =  &(((char*) rq->data)[rq->in]);
		return SUCCESS;
	}
}

StatusCode RingQueue_NotifyInsertionComplete(RingQueue* rq, size_t bytesInserted){
	rq->bytesUsed += bytesInserted;
	rq->in += bytesInserted;
	if (rq->in >= rq->totalSize) {
		if (rq->in > rq->totalSize ) {
			//wrote past the end, so we need to copy some bytes back to the beginning of the queue
			memcpy(rq->data, (void*) &(((char*) rq->data)[rq->totalSize]),(rq->in - rq->totalSize));
			rq->in=(rq->in - rq->totalSize);
		} else {
			rq->in = 0;
		}
	}
	return SUCCESS;
}

StatusCode RingQueue_NextRemovalPoint(RingQueue* rq, char** removalPoint){
	if ((rq->bytesUsed) < RING_QUEUE_REMOVAL_SIZE){
		if(rq->allowUndersizeRemoval){
			*removalPoint = &(((char*) rq->data)[rq->out]);
			return SUCCESS;
		} else {
			*removalPoint = NULL;
			return NOT_READY;
		}
	} else {
		*removalPoint = &(((char*) rq->data)[rq->out]);
		return SUCCESS;
	}
}

StatusCode RingQueue_NotifyRemovalComplete(RingQueue* rq, size_t bytesRemoved){
	if (bytesRemoved>rq->bytesUsed){
		RingQueue_Reset(rq);
	} else {
		rq->bytesUsed -= bytesRemoved;
		rq->out += bytesRemoved;
		if (rq->out >= rq->totalSize) {
			rq->out = 0;
		}
	}
	return SUCCESS;
}

